/* 
   tilt_p5d.c  -  (c) 2000 Till Harbaum

   tilt sensor driver for tilt dongle

   t.harbaum@tu-bs.de
*/

// #include <SysAll.h>
#include <PalmOS.h>
#include "tilt.h"

#include <System/SerialMgrOld.h>
#include <System/SystemMgr.h>

/* interrupt mask register */
#define IMR  (*(unsigned long*)0xfffff304)

/* Dispatch table declarations */
UInt32 install_dispatcher(UInt16, SysLibTblEntryPtr);
MemPtr *gettable(void);

typedef struct {
  Int16 zero_x;
  Int16 zero_y;
  Int16 cpu_type;

  UInt16 LibRef;

  Int16 refcount;
} Lib_globals;

UInt32 start (UInt16 refnum, SysLibTblEntryPtr entryP) {
  UInt32 ret;

  asm("
        movel %%fp@(10),%%sp@-
        movew %%fp@(8),%%sp@-
        jsr install_dispatcher(%%pc)
        movel %%d0,%0
        jmp out(%%pc)
gettable:
        lea jmptable(%%pc),%%a0
        rts

jmptable:
        dc.w 50 | 6n+2 = 50
        dc.w 18 | 2n+4i+2
        dc.w 22
        dc.w 26
        dc.w 30
        dc.w 34
        dc.w 38
        dc.w 42
        dc.w 46
        jmp open(%%pc)
        jmp close(%%pc)
        jmp sleep(%%pc)
        jmp wake(%%pc)
        jmp tilt_zero(%%pc)
        jmp tilt_get(%%pc)
        jmp tilt_get_abs(%%pc)
        jmp unused2(%%pc)
        .asciz \"Tilt Sensor Lib\"
    .even
    out:
        " : "=r" (ret) :);
  return ret;
}

UInt32 install_dispatcher(UInt16 refnum, SysLibTblEntryPtr entryP) {
  MemPtr *table = gettable();
  entryP->dispatchTblP = table;
  entryP->globalsP = NULL;
  return 0;
}

#define CPU_328 (sysFtrNumProcessor328>>16)
#define CPU_EZ  (sysFtrNumProcessorEZ>>16)
#define CPU_VZ  (sysFtrNumProcessorVZ>>16)

UInt16 tilt_check_cpu(void) {
  UInt32 id, chip;
  Err err;

  err = FtrGet(sysFtrCreator, sysFtrNumProcessorID, &id);
  if (err) {
    // can't get that feature; we must be on an old unit
    // without it defined, thus we're on a DragonBall.
    chip = CPU_328;
  } else
    chip = (id & sysFtrNumProcessorMask)>>16;

  return(chip);
}

asm("
.globl getvals_vz
getvals_vz:
        link.w    %a6,#0
	movem.l   %d0-%d5/%a1,-(%a7)
	
        move.l    8(%a6),%a0
	move.l	  #0xfffff439, %a1  | Port J data reg

	move	  #2000,%d0  | init div counter
	move	  %d0,%d1
	move	  %d1,%d2
	move	  %d2,%d3
	move	  %d3,%d4
	move	  %d4,%d5
		
loop_c:	btst.b    #4,(%a1)	| wait while x high
	dbeq	  %d0,loop_c

loop_d:	btst.b    #4,(%a1)	| wait while x low
	dbne	  %d1,loop_d

	| this is time Ta
	
loop_e:	btst.b    #4,(%a1)	| wait while x high
	dbeq	  %d2,loop_e

	| this is time Tb

loop_f:	btst.b    #7,(%a1)	| wait while y high
	dbeq	  %d3,loop_f
		
loop_g:	btst.b    #7,(%a1)	| wait while y low
	dbne	  %d4,loop_g

	| this is time Tc	

loop_h:	btst.b    #7,(%a1)	| wait while y high
	dbeq	  %d5,loop_h

	| this is time Td

	move.w    %d0,(%a0)+
	move.w    %d1,(%a0)+
	move.w    %d2,(%a0)+
	move.w    %d3,(%a0)+
	move.w    %d4,(%a0)+
	move.w    %d5,(%a0)+
						
	movem.l   (%a7)+,%d0-%d5/%a1
	unlk      %a6
	rts
");

asm("
.globl getvals_ez
getvals_ez:
        link.w    %a6,#0
	movem.l   %d0-%d5/%a1,-(%a7)
	
        move.l    8(%a6),%a0
	move.l	  #0xfffff421, %a1  | Port E data reg

	move	  #2000,%d0  | init div counter
	move	  %d0,%d1
	move	  %d1,%d2
	move	  %d2,%d3
	move	  %d3,%d4
	move	  %d4,%d5
		
loop_0:	btst.b    #4,(%a1)	| wait while x high
	dbeq	  %d0,loop_0

loop_1:	btst.b    #4,(%a1)	| wait while x low
	dbne	  %d1,loop_1

	| this is time Ta
	
loop_2:	btst.b    #4,(%a1)	| wait while x high
	dbeq	  %d2,loop_2

	| this is time Tb

loop_3:	btst.b    #7,(%a1)	| wait while y high
	dbeq	  %d3,loop_3
		
loop_4:	btst.b    #7,(%a1)	| wait while y low
	dbne	  %d4,loop_4

	| this is time Tc	

loop_5:	btst.b    #7,(%a1)	| wait while y high
	dbeq	  %d5,loop_5

	| this is time Td

	move.w    %d0,(%a0)+
	move.w    %d1,(%a0)+
	move.w    %d2,(%a0)+
	move.w    %d3,(%a0)+
	move.w    %d4,(%a0)+
	move.w    %d5,(%a0)+
						
	movem.l   (%a7)+,%d0-%d5/%a1
	unlk      %a6
	rts
");

asm("
.globl getvals_328
getvals_328:
        link.w    %a6,#0
	movem.l   %d0-%d5/%a1-%a2,-(%a7)
	
        move.l    8(%a6),%a0
	move.l	  #0xfffff431, %a1  | Port G data reg (RxD)
	move.l	  #0xfffff449, %a2  | Port M data reg (RTS)

	move	  #2000,%d0  | init div counter
	move	  %d0,%d1
	move	  %d1,%d2
	move	  %d2,%d3
	move	  %d3,%d4
	move	  %d4,%d5
		
loop_6:	btst.b    #1,(%a1)	| wait while x high
	dbeq	  %d0,loop_6

loop_7:	btst.b    #1,(%a1)	| wait while x low
	dbne	  %d1,loop_7

	| this is time Ta
	
loop_8:	btst.b    #1,(%a1)	| wait while x high
	dbeq	  %d2,loop_8

	| this is time Tb

loop_9:	btst.b    #0,(%a2)	| wait while y high
	dbeq	  %d3,loop_9
		
loop_a:	btst.b    #0,(%a2)	| wait while y low
	dbne	  %d4,loop_a

	| this is time Tc	

loop_b:	btst.b    #0,(%a2)	| wait while y high
	dbeq	  %d5,loop_b

	| this is time Td

	move.w    %d0,(%a0)+
	move.w    %d1,(%a0)+
	move.w    %d2,(%a0)+
	move.w    %d3,(%a0)+
	move.w    %d4,(%a0)+
	move.w    %d5,(%a0)+
						
	movem.l   (%a7)+,%d0-%d5/%a1-%a2
	unlk      %a6
	rts
");

#define SAMPLE 8
Boolean tilt_measure(Int16 cpu_type, Int16 *x, Int16 *y) {
  unsigned long irq_save;
  unsigned short i,j;
  int ax=0,ay=0;
  int t2[SAMPLE];
  short a[SAMPLE][6];

  /* disable all irqs */
  irq_save = IMR;
  IMR = 0x00ffffff;

  /* read timing values */
  if(cpu_type == CPU_VZ) {
    for(i=0;i<SAMPLE;i++) getvals_vz(a[i]);
  } else if(cpu_type == CPU_EZ) {
    for(i=0;i<SAMPLE;i++) getvals_ez(a[i]);
  } else {
    for(i=0;i<SAMPLE;i++) getvals_328(a[i]);
  }

  /* reenable irqs */
  IMR = irq_save;

  /* counter went down to 0?? -> no chip found */
  if((a[0][0]==-1)||(a[0][1]==-1)) return false;

  for(i=0;i<SAMPLE;i++) {
    for(j=0;j<6;j++)  
      a[i][j] = 2000-a[i][j];

    /* calculate T2 */
    t2[i] = a[i][2]/2 + a[i][3] + a[i][4] + a[i][5]/2;

    ax += ((unsigned long)a[i][2]<<8)/t2[i];
    ay += ((unsigned long)a[i][5]<<8)/t2[i];
  }

  *x=1000-(ax/SAMPLE); 
  *y=1000-(ay/SAMPLE);

  return true;
}

Err open(UInt16 refnum) {
  int i;
  UInt16 cpu;
  Err err;

  /* Get the current globals value */
  SysLibTblEntryPtr entryP = SysLibTblEntry(refnum);
  Lib_globals *gl = (Lib_globals*)entryP->globalsP;

  if (gl) {
    /* We are already open in some other application; just
       increment the refcount */
    gl->refcount++;
    return 0;
  }

  /* We need to allocate space for the globals */
  entryP->globalsP = MemPtrNew(sizeof(Lib_globals));
  MemPtrSetOwner(entryP->globalsP, 0);
  gl = (Lib_globals*)entryP->globalsP;
  
  /* Initialize the globals */
  gl->refcount = 1;
  gl->zero_x   = 0;
  gl->zero_y   = 0;

  /* this driver currently supports 68328, 68EZ328 and 68VZ328 */
  gl->cpu_type = tilt_check_cpu();

  if (SysLibFind("Serial Library", &gl->LibRef)) {
    gl->LibRef = 0;  /* failed ... */
    return TILT_HW_BUSY;
  } else {
    if(err = SerOpen(gl->LibRef, 0, 9600l)) {
      /* failed ... */
      if (err == serErrAlreadyOpen)
        SerClose(gl->LibRef);

      gl->LibRef = 0;
      return TILT_HW_BUSY;
    }
  }

  if((gl->cpu_type != CPU_328)&& 
     (gl->cpu_type != CPU_EZ)&& 
     (gl->cpu_type != CPU_VZ))
      return TILT_WRONG_OS;

  /* initialize sensor */
  if(gl->cpu_type == CPU_VZ) {

    /* RxD, TxD, RTS and CTS do GPIO */
    *((unsigned char*)0xfffff43b) |= 0xf0;

    /* CTS and RxD are inputs, TxD and RTS outputs */
    *((unsigned char*)0xfffff438) = 
      ((*((unsigned char*)0xfffff438)) & ~0x90)|0x60;

    /* power on RTS */
    *((unsigned char*)0xfffff439) &= ~0x60;

  } else if(gl->cpu_type == CPU_EZ) {

    /* RxD, TxD, RTS and CTS do GPIO */
    *((unsigned char*)0xfffff423) |= 0xf0;

    /* CTS and RxD are inputs, TxD and RTS outputs */
    *((unsigned char*)0xfffff420) = 
      ((*((unsigned char*)0xfffff420)) & ~0x90)|0x60;

    /* power on RTS */
    *((unsigned char*)0xfffff421) &= ~0x60;

  } else {

    /* RxD, TxD, RTS and CTS do GPIO */
    *((unsigned char*)0xfffff433) |= 0x03;  /* TxD, RxD */
    *((unsigned char*)0xfffff44B) |= 0x03;  /* RTS, CTS */

    /* CTS and RxD are inputs, TxD and RTS outputs */
    *((unsigned char*)0xfffff430) = 
      ((*((unsigned char*)0xfffff430)) & ~0x02) | 0x01;
    *((unsigned char*)0xfffff448) = 
      ((*((unsigned char*)0xfffff448)) & ~0x01) | 0x02;

    /* power on RTS */
    *((unsigned char*)0xfffff449) &= ~0x02;  
  }

  /* give sensor some time to power up */
  for(i=0;i<10;i++)
    tilt_measure(gl->cpu_type, &gl->zero_x, &gl->zero_y);
  
  /* now try to access sensor */
  if(!tilt_measure(gl->cpu_type, &gl->zero_x, &gl->zero_y)) {

    /* restore old port settings */
    if(gl->cpu_type == CPU_VZ) {
      /* RxD, TxD, RTS and CTS normal operation */
      *((unsigned char*)0xfffff43b) &= ~0xf0;
      /* restore old port settings */
    } else if(gl->cpu_type == CPU_EZ) {
      /* RxD, TxD, RTS and CTS normal operation */
      *((unsigned char*)0xfffff423) &= ~0xf0;
    } else {
      /* RxD, TxD, RTS and CTS normal operation */
      *((unsigned char*)0xfffff433) &= ~0x03;  /* TxD, RxD */
      *((unsigned char*)0xfffff44B) &= ~0x03;  /* RTS, CTS */
    }

    if(gl->LibRef != 0) {
      SerClose(gl->LibRef);
      gl->LibRef = 0;
    }

    return TILT_NO_SENSOR;
  }

  /* sensor is fine */
  return 0;
}

Err close(UInt16 refnum, UInt16 *numappsP) {
  /* Get the current globals value */
  SysLibTblEntryPtr entryP = SysLibTblEntry(refnum);
  Lib_globals *gl = (Lib_globals*)entryP->globalsP;
  if (!gl) return 1;

  /* disable tilt sensor */

  if(gl->cpu_type == CPU_VZ) {
    /* RxD, TxD, RTS and CTS normal operation */
    *((unsigned char*)0xfffff43b) &= ~0xf0;
  } else if(gl->cpu_type == CPU_EZ) {
    /* RxD, TxD, RTS and CTS normal operation */
    *((unsigned char*)0xfffff423) &= ~0xf0;
  } else {
    /* RxD, TxD, RTS and CTS normal operation */
    *((unsigned char*)0xfffff433) &= ~0x03;  /* TxD, RxD */
    *((unsigned char*)0xfffff44B) &= ~0x03;  /* RTS, CTS */
  }

  if(gl->LibRef != 0) {
    SerClose(gl->LibRef);
    gl->LibRef = 0;
  }

  /* Clean up */
  *numappsP = --gl->refcount;
  if (*numappsP == 0) {
    MemChunkFree(entryP->globalsP);
    entryP->globalsP = NULL;
  }
  
  return 0;
}

Err sleep(UInt16 refnum) {
  /* Get the current globals value */
  SysLibTblEntryPtr entryP = SysLibTblEntry(refnum);
  Lib_globals *gl = (Lib_globals*)entryP->globalsP;
  if (!gl) return 1;

  return 0;
}

Err wake(UInt16 refnum) {
  /* Get the current globals value */
  SysLibTblEntryPtr entryP = SysLibTblEntry(refnum);
  Lib_globals *gl = (Lib_globals*)entryP->globalsP;
  if (!gl) return 1;

  return 0;
}

/* zero the tilt sensor */
Err tilt_zero(UInt16 refnum) {
  int i;

  /* Get the current globals value */
  SysLibTblEntryPtr entryP = SysLibTblEntry(refnum);
  Lib_globals *gl = (Lib_globals*)entryP->globalsP;
  if (!gl) return 1; /* We're not open! */  

  /* several times to prevent problems on power up */
  for(i=0;i<10;i++) {
    /* get zero values */
    tilt_measure(gl->cpu_type, &gl->zero_x, &gl->zero_y);
  }

  return 0;
}

/* return normalized tilt value */
Err tilt_get(UInt16 refnum, Int16 *x, Int16 *y) {
  /* Get the current globals value */
  SysLibTblEntryPtr entryP = SysLibTblEntry(refnum);
  Lib_globals *gl = (Lib_globals*)entryP->globalsP;

  if (!gl) return 1; /* We're not open! */

  tilt_measure(gl->cpu_type, x,y);
  *x -= gl->zero_x;
  *y -= gl->zero_y;

  return 0;
}

/* return absolute tilt value */
Err tilt_get_abs(UInt16 refnum, Int16 *x, Int16 *y) {
  /* Get the current globals value */
  SysLibTblEntryPtr entryP = SysLibTblEntry(refnum);
  Lib_globals *gl = (Lib_globals*)entryP->globalsP;

  if (!gl) return 1; /* We're not open! */

  tilt_measure(gl->cpu_type, x,y);

  return 0;
}

/* unused function 2 */
Err unused2(UInt16 refnum) {
  return 0;
}









